import lib.check, lib.curl, json

DOMAIN = 'bitbucket.org'
CONFIG_KEY = 'bitbucket'

ALLOW_FORKS = "allow_forks"
NO_PUBLIC_FORKS = "no_public_forks"
NO_FORKS = "no_forks"
FORK_POLICIES = [ALLOW_FORKS, NO_PUBLIC_FORKS, NO_FORKS]

GIT = "git"
HG = "hg"
SCMS = [GIT, HG]

TRUE = "true"
FALSE = "false"
BOOLEANS = [TRUE, FALSE]

USER = "user"
TEAM = "team"

TARGET = "target"
TARGETS = [USER, TEAM]


def list(target, name, username = None, password = None):

	lib.check.OptionNotInListException(TARGET, target, TARGETS)

	if username is not None :
		username, password = lib.config.prompt_cred(DOMAIN, CONFIG_KEY, username, password)

	urls = {
		USER : "https://bitbucket.org/api/2.0/repositories/%s" % name,
		TEAM : "https://bitbucket.org/api/2.0/teams/%s/repositories" % name
	}

	url = urls[target]

	repositories = []

	while True :

		out, err, p = lib.curl.getjson(url, username = username, password = password, location = True)

		lib.check.SubprocessReturnedFalsyValueException(p.args, p.returncode)

		lib.check.SubprocessOutputEmptyException(p.args, out)

		data = json.loads(out.decode())

		repositories.extend(data["values"])

		if not "next" in data : break

		url = data["next"]

	return repositories


LANGUAGES = [
	None,
	"c",
	"c#",
	"c++",
	"html/css",
	"java",
	"javascript",
	"objective-c",
	"perl",
	"php",
	"python",
	"ruby",
	"shell",
	"sql",
	"other",
	"abap",
	"actionscript",
	"ada",
	"arc",
	"apex",
	"asciidoc",
	"android",
	"asp",
	"arduino",
	"assembly",
	"autoit",
	"blitzmax",
	"boo",
	"ceylon",
	"clojure",
	"coco",
	"coffeescript",
	"coldfusion",
	"common lisp",
	"component pascal",
	"css",
	"cuda",
	"d",
	"dart",
	"delphi",
	"duby",
	"dylan",
	"eiffel",
	"elixir",
	"emacs lisp",
	"erlang",
	"euphoria",
	"f#",
	"fantom",
	"forth",
	"fortran",
	"foxpro",
	"gambas",
	"go",
	"groovy",
	"hack",
	"haskell",
	"haxe",
	"igor pro",
	"inform",
	"io",
	"julia",
	"labview",
	"lasso",
	"latex",
	"limbo",
	"livescript",
	"lua",
	"lilypond",
	"m",
	"markdown",
	"mathematica",
	"matlab",
	"max/msp",
	"mercury",
	"nemerle",
	"nimrod",
	"node.js",
	"nu",
	"object pascal",
	"objective-j",
	"ocaml",
	"occam",
	"occam-π",
	"octave",
	"ooc",
	"other",
	"oxygene",
	"pl/sql",
	"powerbasic",
	"powershell",
	"processing",
	"prolog",
	"puppet",
	"pure basic",
	"pure data",
	"qml",
	"quorum",
	"r",
	"racket",
	"realbasic",
	"restructuredtext",
	"rust",
	"sass/scss",
	"scala",
	"scheme",
	"scilab",
	"sclang",
	"self",
	"smalltalk",
	"sourcepawn",
	"standard ml",
	"supercollider",
	"swift",
	"tcl",
	"tex",
	"typescript",
	"unityscript",
	"unrealscript",
	"vala",
	"verilog",
	"vhdl",
	"viml",
	"visual basic",
	"vb.net",
	"xml",
	"xojo",
	"xpages",
	"xquery",
	"xtend",
	"z shell"
]




import lib.args , os.path

@lib.args.convert( directory = os.path.abspath , port = int )
def lighttpd ( directory , port ) :

	config = """
server.modules = (
		"mod_access",
		"mod_alias",
		"mod_compress",
		"mod_redirect",
		"mod_auth",
		"mod_cgi"
)

cgi.assign      = ( ".cgi" => "" )

server.port                     = %d
server.document-root            = "%s"
server.errorlog                 = "/dev/stdout"
server.dir-listing              = "enable"
dir-listing.encoding            = "utf-8"
index-file.names                = ( "index.cgi", "index.html")
include_shell "/usr/share/lighttpd/create-mime.assign.pl"
"""

	print( config % ( port , directory ) )



import lib.error, lib.list, lib.sys, inspect

def ModuleOrActionNameSpecified (parent, root , inp , i ) :
	if len( inp ) <= i :
		raise lib.error.ModuleOrActionNameNotSpecifiedException( parent , root )

def ModuleOrActionNameExists(parent, root, moduleName, modules):
	if len(modules) == 0:
		raise lib.error.ModuleOrActionNameDoesNotExistException(parent , moduleName, root)

def ModuleOrActionNameNotAmbiguous(parent, moduleName, modules):
	if len(modules) > 1:
		raise lib.error.ModuleOrActionNameAmbiguousException(parent, moduleName, modules)

def KwargsNotSupportedException( actionname, available ):
	if len( available ) == 0:
		raise lib.error.KwargsNotSupportedException( actionname )

def KwargNameExists( kwargname, actionname, matching, available ):
	if len( matching ) == 0:
		raise lib.error.KwargDoesNotExistException( kwargname, actionname, available )

def KwargNameNotAmbiguous( kwargname, actionname, matching ):
	if len( matching ) > 1:
		raise lib.error.KwargNameAmbiguousException( kwargname, actionname, matching )

def NotTooFewArgumentsForAction(moduleName, actionName, n, m, spec):
	if n < m:
		raise lib.error.TooFewArgumentsForActionException(n, spec, actionName, moduleName)

def NotTooManyArgumentsForAction(moduleName, actionName, n, m, spec):
	if spec[1] is None and n > len(spec[0]):
		raise lib.error.TooManyArgumentsForActionException(n, spec, actionName, moduleName)


def OptionNotInListException(key, value, available):
	if value not in available :
		raise lib.error.OptionNotInListException(key, value, available)

def SubprocessReturnedFalsyValueException(cmd, rc):
	if rc != 0 :
		raise lib.error.SubprocessReturnedFalsyValueException(cmd, rc)

def SubprocessOutputEmptyException(cmd, out):
	if out == b"" :
		raise lib.error.SubprocessOutputEmptyException(cmd)

def VersionNotUniqueException(versions):
	if not lib.list.one(versions.values()) :
		raise lib.error.VersionNotUniqueException(versions)

def CannotInferSemverVersionNumberException(old, version):
	if old is None :
		raise lib.error.CannotInferSemverVersionNumberException(version)

def SubprocessArgsEmptyException(args):
	if not args :
		raise lib.error.SubprocessArgsEmptyException()

def SubprocessExecutableNotFoundException(exe):
	if lib.sys.which(exe) is None :
		raise lib.error.SubprocessExecutableNotFoundException(exe)

def ActionDoesNotExistException(A, action, M, module):
	if A is None or not inspect.isfunction(A):
		raise lib.error.ActionDoesNotExistException(action, M, module)

def ModuleDoesNotExistException(M, module, parent):
	if M is None or not inspect.ismodule(M):
		raise lib.error.ModuleDoesNotExistException(module, parent)

def NoneValueException(key, value):
	if value is None :
		raise lib.error.NoneValueException(key)

def FalsyValueException(key, value):
	if not value :
		raise lib.error.FalsyValueException(key)

def NotNoneOrIntegerException(key, value):
	if value is None : return
	try : int(value)
	except : raise lib.error.NotNoneOrIntegerException(key)

try:

	import semantic_version

	def SemverVersionTagNotValidException(version):
		if not semantic_version.validate(version) :
			raise lib.error.SemverVersionTagNotValidException(version)

	def OldSemverVersionTagNotValidException(version, src):
		if not semantic_version.validate(version) :
			raise lib.error.OldSemverVersionTagNotValidException(version, src)

	def NewSemverVersionTagNotGreaterException(a, b):
		if not semantic_version.Version(a) > semantic_version.Version(b) :
			raise lib.error.NewSemverVersionTagNotGreaterException(a, b)

except ImportError as cause:

	e = lib.error.ModuleMissingException(cause, "semantic_version")

	SemverVersionTagNotValidException = lambda version : lib.error.throw(e)
	OldSemverVersionTagNotValidException = lambda version, src : lib.error.throw(e)
	NewSemverVersionTagNotGreaterException = lambda a, b : lib.error.throw(e)





import subprocess


def do(action, *args):
	subprocess.call(['make', action] + args)

def install(*args):
	do('install', *args)

def clean(*args):
	do('clean', *args)

def all(*args):
	do('all', *args)




import lib.sys, json , functools

CMD_CURL = ["curl"]
FLAG_VERBOSE = ["-v"]
PUT = "PUT"
GET = "GET"
POST = "POST"
PATCH = "PATCH"
UPDATE = "UPDATE"
DELETE = "DELETE"


def body(data):
	if data is not None : return ["-d", data]
	else : return []


def resource(contenttype, location, url, accept = None):

	params = ["-H", "Content-Type: %s" % contenttype]

	if accept is not None :
		params += ["-H", "Accept: %s" % accept]

	if location :
		params.append("-L")

	params.append(url)

	return params


def auth(username = None, password = None):
	if username is None : return []
	elif password is None : return ["-u", username]
	else : return ["-u", "%s:%s" % (username, password)]

def request(method):
	return ["-X", method]

def call(method, url, contenttype, data = None, username = None, password = None, location = False, accept = None, **kwargs):

	cmd = []
	cmd.extend(CMD_CURL)
	cmd.extend(request(method))
	cmd.extend(FLAG_VERBOSE)
	cmd.extend(auth(username, password))
	cmd.extend(resource(contenttype, location, url))
	cmd.extend(body(data))

	return lib.sys.call(cmd, **kwargs)


def sendjson(method, url, data = None, username = None, password = None, **kwargs):
	if data is not None : data = json.dumps(data)
	return call(method, url, "application/json", data, username, password, **kwargs)

putjson = functools.partial( sendjson , PUT )
getjson = functools.partial( sendjson , GET )
postjson = functools.partial( sendjson , POST )
updatejson = functools.partial( sendjson , UPDATE )
patchjson = functools.partial( sendjson , PATCH )
deletejson = functools.partial( sendjson , DELETE )




import hashlib, lib.dir, shutil, os, os.path


def read ( f, callback, blocksize = 2**15 ) :
	chunk = f.read(blocksize)
	while len(chunk) > 0:
		callback(chunk)
		chunk = f.read(blocksize)

def hash ( f, h = None, blocksize = 2**15 ) :
	if h is None : h = hashlib.sha256()
	read(f, h.update, blocksize)
	return h

def walk ( src, f ) :

	def callback(path):
		with open(path, 'r') as g : f(g)

	lib.dir.walk(src, f = callback)


def move ( a, cwd ) :

	if not isinstance(a, (list)) : a = [a]

	absolute = lambda path : os.path.join(cwd, path)

	for m in a : shutil.move(*map(absolute, m))


def rm ( *paths, **kwargs ) :

	recursive = kwargs.get( "recursive", False )
	force = kwargs.get( "force", False )

	for path in paths:
		if os.path.isdir(path) : shutil.rmtree(path)
		elif os.path.exists(path) : os.remove(path)
		elif not force : raise Exception("path '%s' does not exist" % path)


def touch ( *files, **kwargs ) :

	"""
		http://stackoverflow.com/a/1160227/1582182
	"""

	times = kwargs.get( "times", None )

	for fname in files :

		with open(fname, 'a'):
			os.utime(fname, times)


def lineiterator ( fp ) :

	return ( line[:-1].decode() for line in iter( fp.readline , b"" ) )




import inspect

def pretty(o, linenos = False, filename = False):
	lines = []
	if filename : lines.append(inspect.getsourcefile(o) + "\n")
	if linenos :
		source, i = inspect.getsourcelines(o)
		n = len(source)
		w = len(str(n))
		for item in source :
			lines.append(str(i).rjust(w) + " " + item)
			i += 1
	else : lines.append(inspect.getsource(o))
	return "".join(lines)




import lib.sys , subprocess

CMD_QSUB = "qsub"
CMD_QDEL = "qdel"
CMD_QSTAT = "qstat"

KEY_JOB_NAME = "Job_Name"
KEY_JOB_STATE = "job_state"
KEY_JOB_OWNER = "Job_Owner"

STATE_COMPLETED = "C"
STATE_QUEUED = "Q"
STATE_RUNNING = "R"

def submit ( name , cmd , nodes = None , cpu = None , output = None , error = None , walltime = None , memory = None , redirect = False ) :

	"""
		qsub python wrapper
	"""

	cp = ""

	if isinstance( nodes , ( list , tuple ) ) :

		cp += ",".join( nodes )

	elif nodes is not None :

		cp += str( nodes )

	if cpu is not None :

		if cp : cp += ":"

		cp += "ppn=" + str( cpu )

	resources = {}

	if cp : resources["nodes"] = cp
	if walltime : resources["walltime"] = walltime
	if memory : resources["mem"] = memory

	resource_list = ",".join(["=".join([key, val]) for key, val in resources.items()])

	submission = [ CMD_QSUB ]

	if resource_list : submission += [ "-l" , resource_list ]

	submission += [ "-N" ,  name ]

	if redirect : submission += [ "-j" , "oe" ]
	if output : submission += [ "-o" ,  output ]
	if error : submission += [ "-e" ,  error ]

	cmd = [ [ "echo" ] + [ cmd ] , submission ]

	return lib.sys.pipeline(*cmd, stdin = subprocess.PIPE, stdout = subprocess.PIPE)



def delete ( queueid ) :

	"""
		qdel python wrapper
	"""

	return lib.sys.call( [ CMD_QDEL , str( queueid ) ] )


def splitowner ( owner ) :

	return owner.split( "@" )


def stat ( ) :

	"""
		qstat python wrapper
	"""

	out, err, p = lib.sys.call( [ CMD_QSTAT , "-f" ] )

	if lib.sys.failure( p ) or not out : return { }

	stat = { }
	current = None

	N_SPACES = 4

	for line in out.decode( ).splitlines( ) :

		if not line : continue

		if line[0] == "\t" :
			current[key] += line[1:]

		elif line[:N_SPACES] == " " * N_SPACES :
			key, part = line[N_SPACES:].split(" = ")
			current[key] = part

		elif line[:8] == "Job Id: ":
			jobid = line[8:]
			stat[jobid] = {}
			current = stat[jobid]

	return stat




import sys

MSG_NOT_VALID = "Please respond with 'yes' or 'no' (or 'y' or 'n')."
YES = "yes"
NO = "no"
VALID = {
	YES: True,
	"y": True,
	"ye": True,
	NO: False,
	"n": False
}

def yesorno(question, default = YES):
	"""Ask a yes/no question via raw_input() and return their answer.

	"question" is a string that is presented to the user.
	"default" is the presumed answer if the user just hits <Enter>.
		It must be "yes" (the default), "no" or None (meaning
		an answer is required of the user).

	The "answer" return value is one of "yes" or "no".
	"""

	if default is None:
		prompt = " [y/n] "
	elif default == YES:
		prompt = " [Y/n] "
	elif default == NO:
		prompt = " [y/N] "
	else:
		raise ValueError("invalid default answer: '%s'" % default)

	while True:
		print(question + prompt, end = '')
		choice = input().lower()
		if default is not None and choice == '':
			return VALID[default]
		elif choice in VALID:
			return VALID[choice]
		else:
			print(MSG_NOT_VALID)




import lib.check, functools, subprocess, os, platform, lib.iterator


STDDEFAULT = "stddefault"


def popen ( cmd , *args , **kwargs ) :

	lib.check.SubprocessArgsEmptyException( cmd )
	lib.check.SubprocessExecutableNotFoundException( cmd[0] )

	if STDDEFAULT in kwargs :
		stddefault = kwargs[STDDEFAULT]
		del kwargs[STDDEFAULT]
	else :
		stddefault = subprocess.PIPE

	Popen = functools.partial(
		subprocess.Popen,
		stdin = stddefault,
		stdout = stddefault,
		stderr = stddefault
	)

	return Popen(cmd, *args, **kwargs)


def call ( cmd , *args , **kwargs ) :

	p = popen( cmd , *args , **kwargs )

	return join( p )

def join(p):
	"""
		Wait for a child's completion.
	"""
	out, err = p.communicate()
	return out, err, p

def extensions():
	if platform.system() == "Windows" :
		return os.environ["PATHEXT"].split(os.pathsep)
	else :
		return [""]


def isexecutable(fpath, exts):
	"""http://stackoverflow.com/a/377028/1582182"""

	for ext in exts:
		path = fpath + ext
		if os.path.isfile(path) and os.access(path, os.X_OK) :
			return True

	return False


def which(program):
	"""http://stackoverflow.com/a/377028/1582182"""
	dirname = os.path.dirname(program)
	exts = extensions()
	if dirname:
		if isexecutable(program, exts) : return program
	else:
		for path in os.environ["PATH"].split(os.pathsep):

			path = path.strip('"')
			exefile = os.path.join(path, program)

			if isexecutable(exefile, exts) : return exefile

	return None

def getlines(cmd, *args, **kwargs):
	"""
		Same as call but only return the output string as a list of lines.
	"""
	out, err, p = lib.sys.call(cmd, *args, **kwargs)
	return out.decode().splitlines()


def whoami():
	"""
		Determines who is the current user.
	"""
	return lib.sys.getlines(["whoami"])[0]


def success(p):
	"""
		Determine if child was successful.
	"""
	return p.returncode == 0

def failure(p):
	"""
		Determine if child failed.
	"""
	return not success(p)


STDIN = "stdin"
STDOUT = "stdout"


def pipeline ( *cmds, **kwargs ) :

	if not cmds : return None, None, 0

	stdin = kwargs.get( STDIN, None )
	stdout = kwargs.get( STDOUT, None )

	inp = stdin
	it = lib.iterator.sentinel( len( cmds ) , subprocess.PIPE , stdout )

	for cmd, out in zip( cmds, it ):
		p = subprocess.Popen( cmd, stdin = inp, stdout = out )
		inp = p.stdout

	out, err = p.communicate()

	return out, err, p



import os, lib.cson, sak.cat, itertools, lib.atom

PROJECTSFILE = os.path.expanduser( "~/.atom/projects.cson" )


def project ( directory ) :

	print( lib.atom.project( directory ) )


def projects () :

	sak.cat.text( PROJECTSFILE )


def add ( directory = ".", *others ) :

	with open( PROJECTSFILE ) as f :
		projects = lib.cson.load( f )


	for target in itertools.chain( [directory], others ) :

		path, name = lib.atom.project( target )

		projects[name] = dict( title = name, paths = [path] )


	with open( PROJECTSFILE, "w" ) as f :
		lib.cson.dump( projects, f )



def remove ( directory = ".", *others ) :

	with open( PROJECTSFILE ) as f :
		projects = lib.cson.load( f )


	for target in itertools.chain( [directory], others ) :

		path, name = lib.atom.project( target )

		if name in projects :
			del projects[name]


	with open( PROJECTSFILE, "w" ) as f :
		lib.cson.dump( projects, f )



import urllib.request

TEXT = "http://oeis.org/search?q=id:%s&fmt=text"
OEIS = "https://oeis.org/%s"

MONTHS = [ "Jan" , "Feb" , "Mar" , "Apr" , "May" , "Jun" , "Jul" , "Aug" , "Sep" , "Oct" , "Nov" , "Dec" ]

def validate ( id ) :

	if len( id ) != 7 :

		print( "ERROR-1: Bad length!" )
		return False

	first , *sequence = id

	if first != "A" :

		print( "ERROR-3: The Sequence does not begin with A." )
		return False

	digits = set( "0123456789" )

	for d in sequence :

		if d not in digits :

			print("ERROR-2: Wrong characters on input string!")
			return False

	return True


def bibtex ( id ) :

	r"""

		>>> from sak.oeis import bibtex
		>>> bibtex( "A00004" )
		ERROR-1: Bad length!
		>>> bibtex( "B000045" )
		ERROR-3: The Sequence does not begin with A.
		>>> bibtex( "A0P0045" )
		ERROR-2: Wrong characters on input string!
		>>> bibtex( "A000045" ) # doctest: +NORMALIZE_WHITESPACE
		@MISC{OEIS:A000045,
			AUTHOR       = {N. J. A. Sloane},
			TITLE        = {The {O}n-{L}ine {E}ncyclopedia of {I}nteger {S}equences},
			HOWPUBLISHED = {\href{https://oeis.org/A000045}{A000045}},
			MONTH        = {Apr},
			YEAR         = {1991},
			NOTE         = {Fibonacci numbers: F(n) = F(n-1) + F(n-2) with F(0) = 0 and F(1) = 1.}
		}
		>>> bibtex( "A000108" ) # doctest: +NORMALIZE_WHITESPACE
		@MISC{OEIS:A000108,
			AUTHOR       = {N. J. A. Sloane},
			TITLE        = {The {O}n-{L}ine {E}ncyclopedia of {I}nteger {S}equences},
			HOWPUBLISHED = {\href{https://oeis.org/A000108}{A000108}},
			MONTH        = {},
			YEAR         = {},
			NOTE         = {Catalan numbers: C(n) = binomial(2n,n)/(n+1) = (2n)!/(n!(n+1)!). Also called Segner numbers.}
		}

	"""

	if not validate( id ) : return

	url = TEXT % id

	try :

		response = urllib.request.urlopen( url )

	except urllib.error.HTTPError as e :

		print( e , ": could not open url %s" % url )
		return

	data = response.read( )
	lines = data.decode( "utf-8" ).splitlines( )

	month , day , year = "" , "" , ""

	for line in lines :

		if line.startswith( "%A" ) :

			author = line[11:]

			for m in MONTHS :

				x = author.rfind( m )

				if x >= 0 :

					month , day , year = author[x:].split( )
					author = author[:x]
					break

			author = ", ".join( x.strip( " _" ) for x in author.strip( ).split("," ) if x )

		if line.startswith( "%N" ) : description = line[11:]

	print( r"""@MISC{OEIS:%s,
	AUTHOR       = {%s},
	TITLE        = {The {O}n-{L}ine {E}ncyclopedia of {I}nteger {S}equences},
	HOWPUBLISHED = {\href{%s}{%s}},
	MONTH        = {%s},
	YEAR         = {%s},
	NOTE         = {%s}
}""" % ( id , author , OEIS % id , id , month , year , description ) )




import lib.http, lib.config

try :
	import urllib.parse as urlparse
	urllib = urlparse
except :
	import urlparse
	import urllib

def resolve(resource, vendors):

	url = urlparse.urlparse(resource)

	p = list(url)

	if url.scheme == '' : p[0] = 'http'

	if url.netloc == '':
		found = False
		for domain, vendor in vendors.items():
			p[1] = domain
			if lib.http.access(urlparse.urlunparse(p)):
				found = True
				break

		if not found:
			return None


	user, _ = urllib.splituser(p[1])
	if user is None:
		vendor = vendors[p[1]]
		user = lib.config.prompt_user(p[1], vendor.__name__)
		p[1] = user + '@' + p[1]

	return urlparse.urlunparse(p)




import time

def callable(fn):
	return hasattr(fn, '__call__')


def val(o, *args, **kwargs):
	if callable(o) : return o(*args, **kwargs)
	else           : return o


def throttle ( times , interval ) :

	def wrap ( fn ) :

		count = 0
		first = time.perf_counter( )

		def wrapper ( *args , **kwargs ) :

			nonlocal count , first

			if count < times :

				count += 1

				return fn( *args , **kwargs )

			else :

				now = time.perf_counter( )

				wait = interval - ( now - first )

				if wait <= 0 :

					count = 1

					first = now

					return fn( *args , **kwargs )

				else :

					time.sleep( wait )

					return wrapper( *args , **kwargs )

		return wrapper

	return wrap





import lib.config, lib.git, lib.hg, lib.error, lib.check, json, lib.sys, lib.bitbucket, lib.input, lib.bytes

DOMAIN = lib.bitbucket.DOMAIN
CONFIG_KEY = lib.bitbucket.CONFIG_KEY
ALLOW_FORKS = lib.bitbucket.ALLOW_FORKS
NO_PUBLIC_FORKS = lib.bitbucket.NO_PUBLIC_FORKS
NO_FORKS = lib.bitbucket.NO_FORKS
FORK_POLICIES = lib.bitbucket.FORK_POLICIES
GIT = lib.bitbucket.GIT
HG = lib.bitbucket.HG
SCMS = lib.bitbucket.SCMS
TRUE = lib.bitbucket.TRUE
FALSE = lib.bitbucket.FALSE
BOOLEANS = lib.bitbucket.BOOLEANS
LANGUAGES = lib.bitbucket.LANGUAGES
USER = lib.bitbucket.USER
TEAM = lib.bitbucket.TEAM


def clone(repo, username = None, scm = GIT):

	url = lib.http.url(DOMAIN, repo, username, secure = True)
	if scm == GIT : lib.git.clone(url)
	else : lib.hg.clone(url)


def new(repository, owner = None, username = None, password = None, is_private = TRUE, scm = GIT, fork_policy = NO_PUBLIC_FORKS, name = None, description = None, language = None, has_issues = FALSE, has_wiki = FALSE):

	lib.check.OptionNotInListException("scm", scm, SCMS)
	lib.check.OptionNotInListException("language", language, LANGUAGES)
	lib.check.OptionNotInListException("has_wiki", has_wiki, BOOLEANS)
	lib.check.OptionNotInListException("has_issues", has_issues, BOOLEANS)
	lib.check.OptionNotInListException("is_private", is_private, BOOLEANS)
	lib.check.OptionNotInListException("fork_policy", fork_policy, FORK_POLICIES)

	username, password = lib.config.prompt_cred(DOMAIN, CONFIG_KEY, username, password)

	if owner is None : owner = username

	parameters = {
		"scm": scm,
		"is_private": is_private,
		"fork_policy": fork_policy,
		"name" : name,
		"description" : description,
		"language" : language,
		"has_issues" : has_issues,
		"has_wiki" : has_wiki
	}

	url = "https://api.bitbucket.org/2.0/repositories/%s/%s" % (owner, repository)

	_, _, p = lib.curl.postjson(url, parameters, username, password, stddefault = None)
	print()
	lib.check.SubprocessReturnedFalsyValueException(p.args, p.returncode)


def group(owner, language, *repositories):

	username, password = lib.config.prompt_cred(DOMAIN, CONFIG_KEY, None, None)

	is_private = TRUE
	scm = GIT
	fork_policy = NO_PUBLIC_FORKS
	name = None
	description = None
	has_issues = FALSE
	has_wiki = FALSE

	for repository in repositories:
		new(repository, owner, username, password, is_private, scm, fork_policy, name, description, language, has_issues, has_wiki)


def list(target, name, username = None, password = None, size = False):
	repos = {}
	for repo in lib.bitbucket.list(target, name, username, password):
		repos[repo["full_name"]] = repo["size"]

	if size :
		total = 0
		for key, val in sorted(repos.items(), key = lambda x : -x[1]):
			print(key, lib.bytes.human(val))
			total += val
		print(lib.bytes.human(total))

	else :
		for key in sorted(repos.keys()) : print(key)


def download ( target, name, username = None, password = None, prompt = True, prefix = "", suffix = "", regexp = "" ):

	for repo in lib.bitbucket.list(target, name, username, password):

		slug = repo["full_name"]

		take = True

		take = take and ( not prefix or slug.startswith( prefix ) )
		take = take and ( not suffix or slug.endswith( prefix ) )
		take = take and ( not regexp or re.match( regexp, slug ) is not None )

		if take and ( not prompt or lib.input.yesorno( "clone '%s'?" % slug ) ) :
			clone( slug, username, scm = repo["scm"] )


def get(owner, repo_slug, username = None, password = None):

	if username is not None :
		username, password = lib.config.prompt_cred(DOMAIN, CONFIG_KEY, username, password)

	url = "https://api.bitbucket.org/2.0/repositories/%s/%s" % (owner, repo_slug)

	out, err, p = lib.curl.getjson(url, username = username, password = password, location = True)

	lib.check.SubprocessReturnedFalsyValueException(p.args, p.returncode)

	return json.loads(out.decode())

def exists(owner, repo_slug, username = None, password = None):

	print("error" not in get(owner, repo_slug, username, password))



import os , lib , lib.sys , os.path

def mount ( directory = '.' , dest = '~/www' ) :

	directory = os.path.abspath( os.path.expanduser( directory ) )
	dest = os.path.abspath( os.path.expanduser( dest ) )

	lib.sys.call( [ 'rm' , dest ] )
	lib.sys.call( [ 'ln' , '-s' , directory , dest ] )


def toggle(which, directory = '.', config_file = 'json/config.json', dry_run = False):
	local = {'root' : os.path.abspath(directory)}
	pre = lambda *args, **kwargs : None

	def callback(helper, config):
		helper.rswitch.toggle(config, local, which)

	if not lib.site.FTPSite.check_local_root(local['root']) : return
	lib.site.FTPSite.wrap(local, config_file, dry_run, pre, callback)


def down(directory = '.', config_file = 'json/config.json', dry_run = False):
	toggle('down', directory, config_file, dry_run)

def up(directory = '.', config_file = 'json/config.json', dry_run = False):
	toggle('up', directory, config_file, dry_run)


def clear ( target , directory = '.' , config_file = 'json/config.json' , dry_run = False ) :

	local = {
		'root' : os.path.abspath( directory ) ,
		'hash' : {} ,
		'tree' : {}
	}

	if not lib.site.FTPSite.check_local_root( local['root'] ) :
		return

	def pre ( helper , config ) :
		pass

	def callback ( helper , config ) :
		helper.remote.recursivermd( target )

	lib.site.FTPSite.wrap( local , config_file , dry_run , pre , callback )


def push(directory = '.', config_file = 'json/config.json', dry_run = False):

	local = {
		'root' : os.path.abspath(directory),
		'hash' : {},
		'tree' : {}
	}

	def pre(helper, config):
		helper.lfetch.local(local['root'], config, local['hash'], local['tree'])

	def callback(helper, config):

		server = {
			'root' : config['root'],
			'hash' : {},
			'tree' : {}
		}

		helper.rfetch.remote(config, server)
		helper.rupdater.all(config, local, server)

	if not lib.site.FTPSite.check_local_root(local['root']) : return
	lib.site.FTPSite.wrap(local, config_file, dry_run, pre, callback)


def diff(directory = '.', config_file = 'json/config.json'):
	return push(directory, config_file, dry_run = True)



def hash(directory = '.', config_file = 'json/config.json'):
	local = {'root' : os.path.abspath(directory)}
	dry_run = False
	pre = lambda *args, **kwargs : None

	def callback(helper, config):

		server = {
			'hash' : {},
			'tree' : {}
		}

		helper.rhasher.hash(config, server['hash'], server['tree'])
		helper.rhasher.send(config, server)

	lib.site.FTPSite.wrap(local, config_file, dry_run, pre, callback)







import inspect, lib.pacman, json


def throw ( e ) :
	raise e

def throws ( MyException ) :
	return lambda *args : throw( MyException( *args ) )

class MainException ( Exception ) :
	def __init__(self, what):
		Exception.__init__(self, what)

	def what(self):
		return self.args[0]

	def __repr__(self):
		return self.what()


class ModuleOrActionNameNotSpecifiedException ( MainException ) :
	def __init__( self , parent , main_module ) :
		fmt = 'Module or action name in module \'%s\' not specified, available modules and actions are : %s'
		args = (parent, ', '.join(sorted(lib.pacman.public(main_module, [inspect.ismodule, inspect.isfunction]))))
		MainException.__init__(self, fmt % args)

class ModuleOrActionNameAmbiguousException ( MainException ) :
	def __init__( self , parent , module_name , matching ) :
		fmt = 'Module or action \'%s\' in module \'%s\' is ambiguous, matching modules and actions are : %s'
		args = (module_name, parent ,', '.join(matching))
		MainException.__init__(self, fmt % args)

class ModuleOrActionNameDoesNotExistException ( MainException ) :
	def __init__(self, parent , module_name, main_module):
		fmt = 'Module or action \'%s\' in module \'%s\' does not exist, available modules and actions are : %s'
		args = (module_name, parent , ', '.join(sorted(lib.pacman.public(main_module, [inspect.ismodule, inspect.isfunction]))))
		MainException.__init__(self, fmt % args)


class KwargsNotSupportedException ( MainException ) :
	def __init__(self, action_name):
		fmt = 'kwargs in action \'%s\' are not supported'
		args = (action_name)
		MainException.__init__(self, fmt % args)

class KwargNameAmbiguousException ( MainException ) :
	def __init__(self, kwarg_name, action_name, matching):
		fmt = 'kwarg \'%s\' in action \'%s\' is ambiguous, matching kwargs are : %s'
		args = (kwarg_name, action_name, ', '.join( matching ))
		MainException.__init__(self, fmt % args)

class KwargDoesNotExistException ( MainException ) :
	def __init__(self, kwarg_name, action_name, available):
		fmt = 'kwarg \'%s\' in action \'%s\' does not exist, available kwargs are : %s'
		args = (kwarg_name, action_name, ', '.join( available ))
		MainException.__init__(self, fmt % args)

class TooFewArgumentsForActionException ( MainException ) :
	def __init__(self, n, spec, action_name, module_name):
		fmt = 'Too few arguments for action \'%s\' in module \'%s\', signature is %s, got %d'
		args = (action_name, module_name, inspect.formatargspec(*spec), n)
		MainException.__init__(self, fmt % args)

class TooManyArgumentsForActionException ( MainException ) :
	def __init__(self, n, spec, action_name, module_name):
		fmt = 'Too many arguments for action \'%s\' in module \'%s\', signature is %s, got %d'
		args = (action_name, module_name, inspect.formatargspec(*spec), n)
		MainException.__init__(self, fmt % args)


class OptionNotInListException ( MainException ) :
	def __init__(self, key, value, available):
		fmt = "%s '%s' is not valid, should be in %s"
		args = (key, value, json.dumps(available))
		MainException.__init__(self, fmt % args)

class OptionOfWrongTypeException ( MainException ) :
	def __init__ ( self, msg ) :
		MainException.__init__( self, msg )


class SubprocessReturnedFalsyValueException ( MainException ) :
	def __init__(self, cmd, rc):
		fmt = "subprocess '%s' exited with return code %d"
		args = (" ".join(cmd), rc)
		MainException.__init__(self, fmt % args)


class SubprocessOutputEmptyException ( MainException ) :
	def __init__(self, cmd):
		fmt = "subprocess '%s' output is 0 bytes long"
		args = (" ".join(cmd))
		MainException.__init__(self, fmt % args)

class SemverVersionTagNotValidException ( MainException ) :
	def __init__(self, version):
		fmt = "version tag '%s' not valid (http://semver.org/)"
		args = version
		MainException.__init__(self, fmt % args)

class OldSemverVersionTagNotValidException ( MainException ) :
	def __init__(self, version, src):
		fmt = "old version tag '%s' in '%s' not valid (http://semver.org/)"
		args = (version, src)
		MainException.__init__(self, fmt % args)

class NewSemverVersionTagNotGreaterException ( MainException ) :
	def __init__(self, a, b):
		fmt = "version tag '%s' should be > than '%s'"
		args = (a, b)
		MainException.__init__(self, fmt % args)



class ModuleMissingException ( MainException ) :
	def __init__(self, cause, name):
		fmt = "'%s' : to fix this --> pip3 install %s"
		args = (cause, name)
		MainException.__init__(self, fmt % args)


class VersionNotUniqueException ( MainException ) :
	def __init__(self, versions):
		fmt = "versions MUST be equal in packages configuration files %s"
		args = json.dumps(versions)
		MainException.__init__(self, fmt % args)

class CannotInferSemverVersionNumberException ( MainException ) :
	def __init__(self, version):
		fmt = "cannot infer version number from '%s' without at least a package configuration file"
		args = version
		MainException.__init__(self, fmt % args)

class SubprocessArgsEmptyException ( MainException ) :
	def __init__(self):
		msg = "error in subprocess : list of arguments is empty"
		MainException.__init__(self, msg)

class SubprocessExecutableNotFoundException ( MainException ) :
	def __init__(self, exe):
		fmt = "executable '%s' not found"
		args = exe
		MainException.__init__(self, fmt % args)


class FileDoesNotExist ( MainException ) :
	def __init__(self, fname):
		fmt = "file '%s' does not exist"
		args = fname
		MainException.__init__(self, fmt % args)

class GithubAPIException ( MainException ) :
	def __init__( self , msg ) :
		MainException.__init__( self , msg )




import functools

# visit http://shields.io/ for spec

DEFAULT = "default"
FLAT = "flat"
FLATSQUARED = "flat-square"

STYLE = [ DEFAULT, FLAT, FLATSQUARED ]

JPG = "jpg"
PNG = "png"
SVG = "svg"

FORMAT = [ JPG, PNG, SVG ]


def imgurl ( path, style = FLAT, fmt = SVG ) :

	if style == DEFAULT :
		return "http://img.shields.io/%s.%s" % ( path, fmt )
	else :
		return "http://img.shields.io/%s.%s?style=%s" % ( path, fmt, style )


def escape ( string ) :

	string = string.replace( "-", "--" )
	string = string.replace( "_", "__" )
	string = string.replace( " ", "_" )

	return string


def custom ( subject, status, color, style = FLAT, fmt = SVG ) :

	subject = escape( subject )
	status = escape( status )
	color = escape( color )

	path = "badge/%s-%s-%s" % ( subject, status, color )

	return imgurl( path, style, fmt )


def service ( *args, **kwargs ) :

	style = kwargs.get( "style", FLAT )
	fmt = kwargs.get( "fmt", SVG )

	args = map( str, args )

	path = '/'.join( args )

	return imgurl( path, style, fmt )


npm = functools.partial( service, "npm" )

npmlicense = functools.partial( npm, "l" )
npmversion = functools.partial( npm, "v" )
npmdownloadsmonth = functools.partial( npm, "dm" )

bower = functools.partial( service, "bower" )
bowerversion = functools.partial( bower, "v" )

github = functools.partial( service, "github" )
githubissues = functools.partial( github, "issues" )
githubtag = functools.partial( github, "tag" )
githubrelease = functools.partial( github, "release" )

scrutinizer = functools.partial( service, "scrutinizer" )
scrutinizerquality = functools.partial( scrutinizer, "g" )
scrutinizercoverage = functools.partial( scrutinizer, "coverage/g" )

travis = functools.partial( service, "travis" )

coveralls = functools.partial( service, "coveralls" )

codeclimate = functools.partial( service, "codeclimate" )
codeclimategithub = functools.partial( codeclimate, "github" )
codeclimatecoverage = functools.partial( codeclimate, "coverage" )
codeclimatecoveragegithub = functools.partial( codeclimatecoverage, "github" )

david = functools.partial( service, "david" )
daviddev = functools.partial( david, "dev" )
davidpeer = functools.partial( david, "peer" )




import lib.json, os.path, lib.fn, lib.github, lib.bitbucket, sak.github, sak.bitbucket



VENDORS = {
	lib.github.DOMAIN : sak.github,
	lib.bitbucket.DOMAIN : sak.bitbucket
}


DEPENDENCIES = ['dep', 'dep-dev']

SAVE = {
	'--save' : DEPENDENCIES[0],
	'--save-dev' : DEPENDENCIES[1]
}

FILENAME_PACKAGE_CONFIGURATION = 'package.json'

DEP = 'dep'


class package(lib.json.proxy):

	def __init__(self, mode = 'r'):
		super().__init__(FILENAME_PACKAGE_CONFIGURATION, mode)



def configOK():
	return os.path.exists(FILENAME_PACKAGE_CONFIGURATION)

def structure():
	if not os.path.exists(DEP):
		os.mkdir(DEP)

	elif not os.path.isdir(DEP):
		print('error -> dep exists and is not a directory')



def depforeach(action, pred, *args, **kwargs):

	with lib.cpm.package() as p:
		for dep in DEPENDENCIES:
			for name, package in p[dep].items():
				path = os.path.join(DEP, name)
				if pred(path):

					_args = []
					_kwargs = {}

					for val in args:
						_args.append(lib.fn.val(val, package))

					for key, val in kwargs:
						_kwargs[key] = lib.fn.val(val, package)


					action(package, *args, **kwargs)



import lib.sys , lib.args

@lib.args.convert( start = int , end = int )
def cut ( source, dest, start, end ) :
	lib.sys.call([
		"pdftk",
		"A=%s" % source,
		"cat",
		"A%d-%d" % (start, end),
		"output",
		dest
	], stddefault = None)


def burst ( source, *others ) :

	lib.sys.call( [
		"pdftk",
		source,
		"burst"
	], stddefault = None )

	if others : burst( *others )


def svg ( source, *others ) :

	lib.sys.call( [
		"pdftocairo",
		"-svg",
		source
	], stddefault = None )

	if others : svg( *others )




import inspect, sys

def inspectable ( fn ) :

	if inspect.isclass( fn ) :

		return fn.__init__

	else :

		return fn


def getfullargspec ( fn ) :

	fn = inspectable( fn )

	return inspect.getfullargspec( fn )




import os, json, socket, tempfile, lib.ftp, lib.nice.ftp, lib.nice.file



class FTPSite(object):

	default = {
		"host" : "hostaddr",
		"username" : "username",
		"root"   : "www",

		"index"  : ".hash",
		"online" : ".online",

		"down"   : ".htaccess",
		"up"     : ".htaccess",

		"tree"   : {},
		"ignore" : []
	}

	def __init__(self, dry_run):
		self.do = not dry_run
		self.ftp = None
		self.remote = None
		self.rhasher = None
		self.rupdater = None
		self.rswitch = None
		self.rfecth = None
		self.lfetch = FTPLocalFetcher()

	def init(self, ftp):
		self.ftp = ftp
		self.remote = lib.nice.ftp.FTP(ftp, self.do)
		self.rhasher = FTPRemoteHasher(self.remote)
		self.rupdater = FTPRemoteUpdater(self.remote)
		self.rswitch = FTPRemoteSwitch(self.remote)
		self.rfetch = FTPRemoteFetcher(self.ftp)

	def check_local_root(root):
		if not os.path.isdir(root):
			print("[Errno 1] Local root '%s' not found" % root)
			return False

		else:
			return True

	def wrap(local, config_file, dry_run, pre, callback):

		try:
			config = FTPSite.default.copy()
			with open(os.path.join(local['root'], config_file), 'r') as f:
				config.update(json.load(f))

		except FileNotFoundError as e:
			print(e)
			return

		helper = FTPSite(dry_run)

		pre(helper, config)

		with lib.ftp.FTP() as ftp:
			helper.init(ftp)

			try:
				ftp.loginprompt(config)
				callback(helper, config)

			except socket.gaierror as e:
				print(e)





class FTPRemoteHasher(object):

	def __init__(self, remote):
		self.remote = remote


	def hash(self, config, htree, tree, current = ''):
		isindex = lambda item : item == config['index']
		ignored = lambda path : path in config['ignore']
		skip = lambda current, item :  isindex(item) or ignored(current + item)
		self.remote.hash(config['root'], htree, tree, skip, current)


	def send(self, config, data):
		path = '/%s/%s' % (config['root'], config['index'])
		self.remote.sendJSON(path, data)
		self.remote.chmod('640', path)






class FTPRemoteUpdater(object):

	def __init__(self, remote):
		self.remote = remote


	def delete_minipaths(self, minipaths, root):
		for minipath in minipaths:
			self.remote.delete('/%s/%s' % (root, minipath))


	def update_moved(self, config, local, h, not_handled, minipaths):
		for minipath in minipaths:

			# moved files
			if minipath not in local['hash'][h]['d']:
				path = '/%s/%s' % (config['root'], minipath)

				if len(not_handled) > 0:
					replace = not_handled[0]
					del not_handled[0]
					self.remote.rename(path, '/%s/%s' % (config['root'], replace))

				else:
					self.remote.delete(path)

	def update_copied(self, config, local, h, not_handled):
		base = local['hash'][h]['s'][0]
		with open('%s/%s' % (local['root'], base), 'rb') as f:
			for i, item in enumerate(not_handled):
				f.seek(0)
				self.remote.storbinary('/%s/%s' % (config['root'], item), f)

	def update_moved_copied_minipaths(self, config, local, h, minipaths):
		not_handled = [x for x in local['hash'][h]['d'] if x not in minipaths]

		self.update_moved(config, local, h, not_handled, minipaths)

		if len(not_handled) > 0 : self.update_copied(config, local, h, not_handled)


	def update_deleted_moved_copied(self, config, local, server):
		for h, minipaths in server['hash'].items():

			# deleted files
			if h not in local['hash']:
				self.delete_minipaths(minipaths, config['root'])

			else:
				self.update_moved_copied_minipaths(config, local, h, minipaths)

	def add_paths(self, config, local, paths):
		for i, path in enumerate(paths['s']):
			with open('%s/%s' % (local['root'], path), 'rb') as f:
				self.remote.storbinary('/%s/%s' % (config['root'], paths['d'][i]), f)

	def update_added(self, config, local, server):
		for h, paths in local['hash'].items():
			# added files
			if h not in server['hash']:
				self.add_paths(config, local, paths)


	def rec_build(local_t, server_t):
		for item, value in local_t.items():
			if type(value) == list:
				server_t[item] = value[0]
			elif type(value) == dict:
				server_t[item] = {}
				FTPRemoteUpdater.rec_build(value, server_t[item])

	def update_index(self, config, local):
		data = {
			'hash' : {},
			'tree' : {}
		}

		for key in local['hash']:
			data['hash'][key] = local['hash'][key]['d']


		FTPRemoteUpdater.rec_build(local['tree'], data['tree'])

		path = '/%s/%s' % (config['root'], config['index'])

		self.remote.sendJSON(path, data)
		self.remote.chmod('640', path)

	def all(self, config, local, server):
		self.remote.makedirs(config['root'], local['tree'], server['tree'])
		self.update_deleted_moved_copied(config, local, server)
		self.update_added(config, local, server)
		self.update_index(config, local)
		self.remote.removedirs(config['root'], local['tree'], server['tree'])






class FTPRemoteSwitch(object):

	def __init__(self, remote):
		self.remote = remote


	def toggle(self, config, local, which):
		src = os.path.join(local['root'], config[which])
		if os.path.isfile('%s%s' % (src, config['online'])) : src += config['online']

		with open(src, 'rb') as f:
			self.remote.storbinary('/%s/%s' % (config['root'], config['up']), f)


	def down(self, config, local):
		return self.server_switch(config, local, 'down')


	def up(self, config, local):
		return self.server_switch(config, local, 'up')





class FTPLocalFetcher(object):

	def __init__(self):
		pass

	def local_file(online, hash_t, tree_i, tree, dir_list, item, minipath, path):
		base, ext = os.path.splitext(minipath)
		if ext == online :
			if os.path.basename(minipath) in tree : return
			else : dest = base
		elif item + online in dir_list :
			dest = minipath
			path += online
			minipath += online
		else : dest = minipath
		h_ascii = lib.nice.file.hascii(path)
		hash_t.setdefault(h_ascii, {'s' : [], 'd' : []})
		hash_t[h_ascii]['s'].append(minipath)
		hash_t[h_ascii]['d'].append(dest)
		tree_i[item] = [h_ascii, dest]

	def local(self, root, config, hash_t, tree_i, current = '', tree = None):
		if tree is None : tree = config['tree']
		dir_list = os.listdir(root + '/' + current)
		for item, what in tree.items():
			minipath = current + item
			path = root + '/' + minipath
			if os.path.isfile(path):
				FTPLocalFetcher.local_file(config['online'], hash_t, tree_i, tree, dir_list, item, minipath, path)

			elif os.path.isdir( path ) :

				if what is None :
					what = dict( ( sub, None ) for sub in os.listdir( path ) )

				tree_i[item] = {}
				self.local(root, config, hash_t, tree_i[item], current + item + '/', what)

class FTPRemoteFetcher(object):

	def __init__(self, ftp):
		self.ftp = ftp

	def remote(self, config, server):
		index_file = '/%s/%s' % (config['root'], config['index'])
		if self.ftp.isfile(index_file):
			chuncks = []
			self.ftp.retrlines('RETR %s' % index_file, chuncks.append)
			index = ''.join(chuncks)
			data = json.loads(index)
			server['hash'] = data['hash']
			server['tree'] = data['tree']






class ansy(object):
	HEADER = '\033[95m'
	OKBLUE = '\033[94m'
	OKGREEN = '\033[92m'
	WARNING = '\033[93m'
	FAIL = '\033[91m'
	ENDC = '\033[0m'
	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'

	def disable(self):
		self.HEADER = ''
		self.OKBLUE = ''
		self.OKGREEN = ''
		self.WARNING = ''
		self.FAIL = ''
		self.ENDC = ''
		self.BOLD = ''
		self.UNDERLINE = ''



import lib.color, lib.css, subprocess

def build(src, ugly = None):
	cmd = ['compass', 'compile', src]
	if ugly is not None : cmd += ['-e', 'production', '--force']
	subprocess.call(cmd)

def watch(src):
	cmd = ['compass', 'watch', src]
	subprocess.call(cmd)

def hex2rgba(h):
	print(lib.css.rgba2str(*lib.css.hex2rgb(h)))

def hex2rgb(h):
	print(lib.css.rgb2str(*lib.css.hex2rgb(h)[0:3]))

def rgb2hex(rgb):
	print(lib.css.rgb2hex(*lib.css.str2rgb(rgb)))

def rgba2hex(rgb):
	print(lib.css.rgba2hex(*lib.css.str2rgb(rgb)))

def darken(inp, p):
	helper.light(inp, p, lib.color.darken)

def lighten(inp, p):
	helper.light(inp, p, lib.color.lighten)

def rgb2hsl(rgb):
	print(lib.css.hsl2str(*lib.color.rgb2hsl(*lib.css.str2rgb(rgb))))

def rgba2hsla(rgb):
	print(lib.css.hsla2str(*lib.color.rgba2hsla(*lib.css.str2rgb(rgb))))

def hsl2rgb(hsl):
	print(lib.css.rgb2str(*lib.color.hsl2rgb(*lib.css.str2hsl(hsl))))

def hsla2rgba(hsl):
	print(lib.css.rgba2str(*lib.color.hsla2rgba(*lib.css.str2hsl(hsl))))


class helper(object):
	def light(inp, p, transform):

		if inp[0] == 'r':
			out = lambda *x : lib.css.rgb2str(*lib.color.hsl2rgb(*x))
			hsl = lib.color.rgb2hsl(*lib.css.str2rgb(inp))
		elif inp[0] == 'h':
			out = lib.css.hsl2str
			hsl = lib.css.str2hsl(inp)
		else:
			out = lambda *x : lib.css.rgb2hex(*lib.color.hsl2rgb(*x))
			hsl = lib.color.rgb2hsl(*lib.css.hex2rgb(inp)[0:3])

		print(out(*transform(*hsl, p = int(p))))



import os, sys, lib.file, lib.check, fileinput, lib.args

def concat(src, out):

	with open(out, 'w') as f:

		def callback(g):
			lib.file.read(g, f.write)
			f.write(os.linesep)

		lib.file.walk(src, callback)


def append(out, txt):

	with open(out, 'a') as f:
		f.write(txt)
		f.write(os.linesep)


TAB = "tab"
SPACE = "space"

INDENT_MODES = [TAB, SPACE]

INDENT_DEFAULT_WIDTH = {
	TAB : 1,
	SPACE : 4
}

INDENT_CHAR = {
	TAB : '\t',
	SPACE : ' '
}

def indent(src = sys.stdin, mode = TAB, width = None):

	lib.check.OptionNotInListException("mode", mode, INDENT_MODES)

	c = INDENT_CHAR[mode]

	if width is None : width = INDENT_DEFAULT_WIDTH[mode]

	lineprepend(src, c, width)


def lineprepend(src = sys.stdin, string = "", width = 1):

	width = int(width)
	sequence = string * width

	if src != sys.stdin :
		src = open(src, 'r')

	for line in src:
		print(sequence + line, end = '')

	if src != sys.stdin :
		src.close()


def uppercase ( *args ) :

	for line in fileinput.input( args ) :

		print( line.upper() , end = "" )


def lowercase ( *args ) :

	for line in fileinput.input( args ) :

		print( line.lower() , end = "" )

@lib.args.mandatory( n = True )
@lib.args.convert( n = int )
def truncate ( *args , n = None ) :

	for line in fileinput.input( args ) :

		print( line[:n] , end = "" )

def split ( *args ) :

	for line in fileinput.input( args ) :

		print( *line.split( ) , sep = "\n" , end = "\n" )




import os, json, lib.json

def projectjson(path):
	return {
		'folders' : [
			{
				'path' : path,
				'folder_exclude_patterns': []
			}
		]
	}

def projectpath(directory):
	root = os.path.abspath(directory)
	name = os.path.basename(root)
	return root, os.path.join(root, name + '.sublime-project')

def workspacepath(directory):
	root = os.path.abspath(directory)
	name = os.path.basename(root)
	return root, os.path.join(root, name + '.sublime-workspace')

def configdir(v):
	return os.path.expanduser('~/.config/sublime-text-') + str(v)

def configdirs():
	for v in [2, 3] :
		d = configdir(v)
		if os.path.isdir(d): yield d

def sessionfile(d):
	return os.path.join(d, "Settings", "Session.sublime_session")


def add(directory = '.', *others):
	path, fproject = projectpath(directory)

	with open(fproject, 'w') as f:
		project = projectjson(path)
		json.dump(project, f, indent = '\t')

	for d in configdirs():

		fsession = sessionfile(d)

		with lib.json.proxy(fsession, "w", strict = False, indent = '\t', throws = True) as config :
			config['workspaces']['recent_workspaces'].insert(0, fproject)

	if others : add(*others)

def remove(directory = '.', *others):
	path, fproject = projectpath(directory)

	if os.path.isfile(fproject) : os.remove(fproject)
	else : print("could not find '%s'" % fproject)

	path, fworkspace = workspacepath(directory)
	if os.path.isfile(fworkspace) : os.remove(fworkspace)

	for d in configdirs():

		fsession = sessionfile(d)

		with lib.json.proxy(fsession, "w", strict = False, indent = '\t', throws = True) as config :
			recent = config['workspaces']['recent_workspaces']
			if fproject in recent : recent.remove(fproject)

	if others : remove(*others)




import time , datetime

def nanoseconds ( ) :
	return int( round( time.time( ) * 1e9 ) )

def parsedmy ( dmy ) :

	datefmt = "%d/%m/%Y"

	parsed = datetime.datetime.strptime( dmy , datefmt )

	timetuple = parsed.timetuple( )

	return time.mktime( timetuple ) * 1e9

def pretty ( timestamp ) :
	return datetime.datetime.utcfromtimestamp( timestamp / 1e9 ).strftime( "%Y-%m-%d %H:%M:%S" )



import lib.config, lib.json, lib.file

def gen():
	with open(lib.config.fname, 'w') as f : lib.json.pretty(lib.config.new(), f)

def cat():
	with open(lib.config.fname, 'r') as f : lib.file.read(f, print)

def add(user, passwd):
	with lib.config.proxy( 'w' ) as config : config['u'][user] = passwd

def rm(user):
	with lib.config.proxy( 'w' ) as config : config['u'].pop(user)

def link(module, user):
	with lib.config.proxy( 'w' ) as config : config['m'][module] = user

def unlink(module):
	with lib.config.proxy( 'w' ) as config : config['m'].pop(module)

def user(user):
	print(lib.config.user(user))

def module(module):
	print(lib.config.module(module))




import re

_v = 'aeiou'
_V = _v.upper()
_vV = _v + _V

_c = 'zrtypqsdfghjklmwxcvbn'
_C = _c.upper()
_cC = _c + _C


def filt ( s, f ) :
	return ''.join( [c for c in s if c in f] )

vowe  = lambda s: filt( s, _vV )
cons  = lambda s: filt( s, _cC )
vowel = lambda s: filt( s, _v )
consl = lambda s: filt( s, _c )
voweu = lambda s: filt( s, _V )
consu = lambda s: filt( s, _C )


def natural ( s ) :
	return [ int( c ) if c.isdigit() else c for c in re.split( "([0-9]+)", s ) ]

def inside ( a , b ) :
	"""

		Checks if a is inside b, in order.

		>>> from lib.str import *
		>>> inside( "kw" , "keywords" )
		True
		>>> inside( "kywo" , "keywords" )
		True
		>>> inside( "ka" , "keywords" )
		False
		>>> inside( "keeywords" , "keywords" )
		False
		>>> inside( "a" , "action" )
		True
		>>> inside( "a" , "throttled" )
		False

	"""

	i = 0
	j = 0
	m = len( a )
	n = len( b )

	while i < m :

		if n - j < m - i : return False

		if a[i] == b[j] : i += 1

		j += 1

	return True


def utod ( s ) :
	"""
		Replace underscores with dashes.

		>>> from lib.str import *
		>>> utod( "_abc_d_e_" )
		'-abc-d-e-'

	"""
	return s.replace( '_' , '-' )


NAME_RESOLVER = (
	# MATCH
	lambda stringset, alter, string : [ x for x in stringset if alter( x ) == string ] ,
	# PREFIX
	lambda stringset, alter, string : [ x for x in stringset if alter( x ).startswith( string ) ] ,
	# SUFFIX
	lambda stringset, alter, string : [ x for x in stringset if alter( x ).endswith( string ) ] ,
	# SUBSTR
	lambda stringset, alter, string : [ x for x in stringset if string in alter( x ) ]
)

NAME_RESOLVER2 = (
	# INSIDE
	lambda stringset, alter, string : [ x for x in stringset if inside( string , alter( x ) ) ] ,
)


NAME_TRANSFORMER = (
	lambda string : utod( string ) ,
	lambda string : utod( string.lower( ) ) ,
	lambda string : cons( string ) ,
	lambda string : cons( string.lower( ) )
)

NAME_TRANSFORMER2 = (
	NAME_TRANSFORMER[2] , NAME_TRANSFORMER[3] ,
	NAME_TRANSFORMER[0] , NAME_TRANSFORMER[1]
)

NAME_INSPECTORS = (
	(  NAME_TRANSFORMER ,  NAME_RESOLVER ) ,
	( NAME_TRANSFORMER2 , NAME_RESOLVER2 )
)


def mostlikely ( target , stringset ) :

	"""

		Determines which strings out of stringset are most likely to be
		targeted by the target string.

		>>> from lib.str import *
		>>> mostlikely( 'a' , [ 'action' , 'throttled' ] )
		['action']
		>>> mostlikely( '' , [ 'a' , 'b' , 'c' ] )
		[]
		>>> mostlikely( 'kw' , [ 'kwargs' , 'keywords' ] )
		['kwargs']
		>>> mostlikely( 'k' , [ 'kwargs' , 'keywords' ] )
		['kwargs', 'keywords']
		>>> mostlikely( 'kw' , [ 'username' , 'keywords' ] )
		['keywords']
		>>> mostlikely( 'x-y' , [ 'X_Y' , 'x_y' , 'XY' , 'xy' ] )
		['x_y']
		>>> mostlikely( 'X-Y' , [ 'X_Y' , 'x_y' , 'XY' , 'xy' ] )
		['X_Y']
		>>> mostlikely( 'xy' , [ 'X_Y' , 'x_y' , 'XY' , 'xy' ] )
		['xy']
		>>> mostlikely( 'XY' , [ 'X_Y' , 'x_y' , 'XY' , 'xy' ] )
		['XY']
		>>> mostlikely( 'x-y' , [ 'XY' , 'xy' ] )
		['xy']
		>>> mostlikely( 'X-Y' , [ 'XY' , 'xy' ] )
		['XY']
		>>> mostlikely( 'xy' , [ 'X_Y' , 'x_y' ] )
		['x_y']
		>>> mostlikely( 'XY' , [ 'X_Y' , 'x_y' ] )
		['X_Y']

	"""

	for transformers , resolvers in NAME_INSPECTORS :

		for alter in transformers :

			string = alter( target )

			# if alter reduces target to the empty
			# string do not forget to skip
			if not string : continue

			for resolver in resolvers :
				matching = resolver( stringset, alter, string )

				if len( matching ) > 0 :
					return matching

	return []




import lib.error

PREFIX = 'v'
KEYS = ['major', 'minor', 'patch']

try:

	import semantic_version

	def resolve(base, key):
		v = semantic_version.Version(base)
		setattr(v, key, getattr(v, key) + 1)
		i = KEYS.index(key) + 1
		for k in KEYS[i:] : setattr(v, k, 0)
		return str(v)


except ImportError as cause:

	e = lib.error.ModuleMissingException(cause, "semantic_version")

	resolve = lambda base, key : lib.error.throw(e)


def isspecial(v):
	return v in KEYS



import shutil, sak.github, lib.github, lib.sak, sak.npm
import lib.bower, lib.check, collections, lib.dir, lib.file
import lib.codebricks, fileinput, lib.args, lib.http
import os

TRAVISCI = lib.codebricks.TRAVISCI
DRONEIO = lib.codebricks.DRONEIO
CI = lib.codebricks.CI
FLAT = lib.codebricks.FLAT
SVG = lib.codebricks.SVG
README = "README.md"

def new ( name, subject, keywords = None, ci = TRAVISCI, username = None, password = None ) :

	lib.check.OptionNotInListException("ci", ci, CI)

	username, password = lib.github.credentials(username, password)

	repo = "js-" + name

	description = "%s code bricks for JavaScript" % subject

	fmtargs = dict(name = name, description = description, repo = repo, username = username)

	homepage = "http://%(username)s.github.io/%(repo)s/" % fmtargs
	githubpage = "https://github.com/%(username)s/%(repo)s" % fmtargs
	issuespage = githubpage + "/issues"

	keywords = lib.args.listify( keywords )

	keywords = sorted(list(set(["js", "javascript", "bricks"] + keywords)))

	license = dict(name = "AGPL-3.0", template = "agpl-3.0")

	qualifiedname = "%(username)s-%(repo)s" % fmtargs

	fmtargs["qualifiedname"] = qualifiedname

	sak.github.new(
		repo,
		username = username,
		password = password,
		auto_init = lib.github.TRUE,
		private = lib.github.FALSE,
		description = description,
		homepage = homepage,
		has_issues = lib.github.TRUE,
		has_wiki = lib.github.TRUE,
		has_downloads = lib.github.TRUE,
		gitignore_template = "Node",
		license_template = license["template"]
	)

	_, _, p = sak.github.clone( "%(username)s/%(repo)s" % fmtargs, username = username )


	with lib.dir.cd(repo) :

		jsonhook = collections.OrderedDict

		with lib.json.proxy(".groc.json", "w", object_pairs_hook = jsonhook) as groc :
			groc["glob"] = ["js/src/**/*.js", README]
			groc["github"] = True

		with open(".gitignore", "a") as gitignore :
			gitignore.write("\n")
			gitignore.write("# groc\n")
			gitignore.write("doc\n")

		with open(README, "w") as readme :
			readme.write("[%(repo)s](http://%(username)s.github.io/%(repo)s)\n" % fmtargs)
			readme.write("==\n")
			readme.write("\n")
			readme.write("%(description)s\n" % fmtargs)
			readme.write("\n")
			lib.codebricks.badges( username, repo, ci, lambda s : readme.write(s + "\n") )
			readme.write("\n")
			readme.write( lib.codebricks.installationinstructions( username , name ) )

		with lib.json.proxy("package.json", "w", object_pairs_hook = jsonhook) as npm :
			npm["name"] = qualifiedname
			npm["version"] = "0.0.0"
			npm["description"] = description
			npm["main"] = "js/dist/%(name)s.js" % fmtargs
			npm["dependencies"] = {}
			npm["devDependencies"] = {"aureooms-node-package": "^4.1.0"}
			npm["scripts"] = {}
			npm["scripts"]["build"] = "./node_modules/.bin/aureooms-node-package-build"
			npm["scripts"]["test"] = "./node_modules/.bin/aureooms-node-package-test"
			npm["scripts"]["doc"] = "./node_modules/.bin/groc"
			npm["repository"] = {}
			npm["repository"]["type"] = "git"
			npm["repository"]["url"] = "https://github.com/%(username)s/%(repo)s.git" % fmtargs
			npm["keywords"] = keywords
			npm["author"] = username
			npm["license"] = license["name"]
			npm["bugs"] = {}
			npm["bugs"]["url"] = issuespage
			npm["homepage"] = homepage


		with lib.json.proxy("bower.json", "w", object_pairs_hook = jsonhook) as bower :
			bower["name"] = qualifiedname
			bower["version"] = "0.0.0"
			bower["description"] = description
			bower["main"] = "js/dist/%(name)s.js" % fmtargs
			bower["ignore"] = [
				"js/index.js",
				"js/src",
				"test",
				"pkg.json",
				"package.json",
				".groc.json",
				".travis.yml",
				".gitignore",
				README
			]
			bower["license"] = license["name"]
			bower["homepage"] = homepage

		with lib.json.proxy("component.json", "w", object_pairs_hook = jsonhook) as component :
			component["name"] = qualifiedname
			component["repo"] = "%(username)s/%(repo)s" % fmtargs
			component["version"] = "0.0.0"
			component["license"] = license["name"]
			component["description"] = description
			component["main"] = "js/dist/%(name)s.js" % fmtargs
			component["scripts"] = ["js/dist/%(name)s.js" % fmtargs]

		with lib.json.proxy("pkg.json", "w", object_pairs_hook = jsonhook) as pkg :
			pkg["name"] = name
			pkg["src"] = "js/src/"
			pkg["out"] = "js/dist/"
			pkg["code"] = {}
			pkg["code"]["main"] = ["js", "dist", "%(name)s.js" % fmtargs]
			pkg["code"]["test"] = ["test", "js"]
			pkg["debug"] = False


		lib.dir.makedirs("js/src", "test/js/src")
		lib.file.touch("js/src/dummy.js")

		shutil.copy(lib.sak.data("codebricks", "js-index.js"), "js/index.js")
		shutil.copy(lib.sak.data("codebricks", "test-js-index.js"), "test/js/index.js")
		shutil.copy(lib.sak.data("codebricks", "test-js-src-dummy.js"), "test/js/src/dummy.js")

		if ci == TRAVISCI :
			shutil.copy(lib.sak.data("codebricks", ".travis.yml"), ".travis.yml")

		lib.git.add("--all", ".")
		lib.git.commit("-am", "$ codebricks new")
		lib.git.push()
		sak.npm.install()
		sak.npm.release("0.0.1")
		lib.bower.register(qualifiedname, "github.com/%(username)s/%(repo)s" % fmtargs, force = True)


def badges ( username, repo, ci = TRAVISCI, style = FLAT, fmt = SVG ) :

	lib.codebricks.badges( username, repo, ci, print, style, fmt )


def fork ( oldrepo, name, subject, keywords = None, ci = TRAVISCI, username = None, password = None ) :

	username, password = lib.github.credentials( username, password )

	oldowner, oldslug = oldrepo.split('/')

	slug = "js-" + name

	qualifiedname = "%s-%s" % ( username, slug )

	description = "%s code bricks for JavaScript" % subject

	homepage = "http://%s.github.io/%s/" % ( username, slug )
	githubpage = "https://github.com/%s/%s" % ( username, slug )
	issuespage = githubpage + "/issues"

	keywords = lib.args.listify( keywords )

	keywords = sorted(list(set(["js", "javascript", "bricks"] + keywords)))

	sak.github.new(
		slug,
		username = username,
		password = password,
		auto_init = lib.github.FALSE,
		private = lib.github.FALSE,
		description = description,
		homepage = homepage,
		has_issues = lib.github.TRUE,
		has_wiki = lib.github.TRUE,
		has_downloads = lib.github.TRUE
	)

	sak.github.clone( oldrepo , dest = slug , username = username )

	with lib.dir.cd( slug ) :

		jsonhook = collections.OrderedDict

		for line in fileinput.input( README, inplace = True ) :
			line = line.replace( oldslug, slug )
			line = line.replace( oldowner, username )
			print( line, end = "" )

		with open( README, "a" ) as readme :
			readme.write( "\n" )
			readme.write( "***( forked from [%s](https://github.com/%s) )***" % ( oldslug, oldrepo ) )
			readme.write( "\n" )

		with lib.json.proxy( "package.json", "w", object_pairs_hook = jsonhook ) as npm :
			npm["name"] = qualifiedname
			npm["description"] = description
			npm["main"] = "js/dist/%s.js" % name
			npm["repository"]["url"] = "https://github.com/%s/%s.git" % ( username, slug )
			npm["keywords"] = keywords
			npm["author"] = username
			npm["bugs"]["url"] = issuespage
			npm["homepage"] = homepage

		with lib.json.proxy( "bower.json", "w", object_pairs_hook = jsonhook ) as bower :
			bower["name"] = qualifiedname
			bower["description"] = description
			bower["main"] = "js/dist/%s.js" % name
			bower["homepage"] = homepage

		with lib.json.proxy("component.json", "w", object_pairs_hook = jsonhook) as component :
			component["name"] = qualifiedname
			component["repo"] = "%s/%s" % ( username, slug )
			component["description"] = description
			component["main"] = "js/dist/%s.js" % name
			component["scripts"] = ["js/dist/%s.js" % name]

		with lib.json.proxy( "pkg.json", "w", object_pairs_hook = jsonhook ) as pkg :
			pkg["name"] = name
			pkg["code"]["main"] = ["js", "dist", "%s.js" % name]


		lib.file.rm( "js/dist" )

		url = lib.http.url( "github.com" , path = "%s/%s" % ( username , slug ) , username = username , secure = True )

		lib.git.remote( "set-url", "origin", url )
		lib.git.add( "--all", "." )
		lib.git.commit( "-am", "$ codebricks fork %s" % oldrepo )
		lib.git.push( "-u", "origin", "master" )
		sak.npm.install()
		sak.npm.release( "major" )
		lib.bower.register( qualifiedname, "github.com/%s/%s" % ( username, slug ), force = True )


def component ( qualifiedname , repo , name , version , license , description ) :

	with lib.json.proxy( "component.json" , "w" , object_pairs_hook = collections.OrderedDict ) as component :
		component["name"] = qualifiedname
		component["repo"] = repo
		component["version"] = version
		component["license"] = license
		component["description"] = description
		component["main"] = "js/dist/%s.js" % name
		component["scripts"] = ["js/dist/%s.js" % name]

def usecomponent ( *dirs ) :

	for d in dirs :

		if "component.json" in os.listdir( d ) : continue

		with lib.dir.cd( d ) :

			with lib.json.proxy( "pkg.json" ) as pkg : pass

			with lib.json.proxy( "bower.json" ) as bower : pass

			name = pkg["ns"] if "ns" in pkg else pkg["name"]

			qualifiedname = bower["name"]

			author , *_ = qualifiedname.split("-")

			repo = author + "/js-" + name

			version = bower["version"]

			license = bower["license"]

			description = bower["description"]

			component ( qualifiedname , repo , name , version , license , description )

			lib.git.add( "--all" , "." )
			lib.git.commit( "-am", "add component.json" )
			lib.git.push( )
			sak.npm.release( "patch" )

def installationinstructions ( username , name ) :

	"""

		>>> import sak.codebricks
		>>> sak.codebricks.installationinstructions( "abcd" , "efgh" )
		Can be managed through [duo](https://github.com/duojs/duo),
		[component](https://github.com/componentjs/component),
		[bower](https://github.com/bower/bower), or
		[npm](https://github.com/npm/npm).
		<BLANKLINE>
		```js
		let efgh = require( "abcd-js-efgh" ) ;
		```

	"""

	print( lib.codebricks.installationinstructions( username , name ) )



import os, subprocess

def x2eps(img, out = None, width = None, height = None):
	inkscape.convert(img, out, width, height, '-E', '.eps')

def x2png(img, out = None, width = None, height = None):
	inkscape.convert(img, out, width, height, '-e', '.png')

def x2pdf(img, out = None, width = None, height = None):
	inkscape.convert(img, out, width, height, '-A', '.pdf')

def x2svg(img, out = None, width = None, height = None):
	inkscape.convert(img, out, width, height, '-l', '.svg')

def resize(img, out, width, height, mode = '^'):
	subprocess.call(['convert', img, '-resize', '%sx%s%s' % (width, height, mode), out])

def crop(img, out, width, height, gravity = 'Center', left = '0', top = '0'):
	subprocess.call(['convert', img, '-gravity', gravity, '-crop', '%sx%s+%s+%s' % (width, height, left, top), out])


class inkscape(object):
	def convert(img, out, width, height, t, ext):

		if out is None:
			base, _ = os.path.splitext(img)
			out = base + ext

		args = ['inkscape', '-z', '-T', t, out, img]

		if width  is not None : args.extend(['--export-width' , width ])
		if height is not None : args.extend(['--export-height', height])

		subprocess.call(['inkscape', '-z', '-T', t, out, img])



import os, lib, getpass, functools

try :
	import urllib.parse as urllib
except :
	import urllib

fname = os.path.expanduser('~/.sak')

def new():
	return {'u':{}, 'm':{}}

def proxy ( mode = 'r', default = None, throws = False, **kwargs ) :

	if default is None :
		default = new()

	return lib.json.proxy( fname, mode, default, throws, **kwargs )

def user(user):

	with lib.config.proxy() as config :
		passwd = config['u'].get(user, None)

	return passwd

def module(module):

	with lib.config.proxy() as config :
		user = config['m'].get(module, None)
		passwd = config['u'].get(user, None)

	return user, passwd



def prompt_cred(host, module, user = None, passwd = None, scheme = "https", port = 80):

	user = prompt_user(host, module, user, passwd, scheme, port)
	passwd = prompt_passwd(host, module, user, passwd, scheme, port)

	return user, passwd

def prompt_user(host, module, user = None, passwd = None, scheme = "https", port = 80):

	url = '%s://%s:%d' % (scheme, host, port)

	if user is None : user, passwd = lib.config.module(module)
	if user is None : user = input('Username for \'%s\' : ' % url)

	return user


def prompt_passwd(host, module, user, passwd = None, scheme = "https", port = 80):

	url = '%s://%s@%s:%d' % (scheme, urllib.quote_plus(user), host, port)

	if passwd is None : passwd = lib.config.user(user)
	if passwd is None : passwd = getpass.getpass('Password for \'%s\' : ' % url)

	return passwd




def darken(h, s, l, p):
	return lighten(h, s, l, -p)

def lighten(h, s, l, p):
	return h, s, l + p * 2.55

def rgb2hsl(r, g, b):
	M = max(r, g, b)
	m = min(r, g, b)
	C = M - m
	l = (M + m) / 2
	s = 0 if C == 0 else C / (255 - abs(2*l - 255))
	if   C == 0 : h = 0
	elif r == M : h = (g - b) / C
	elif g == M : h = (b - r) / C + 2
	elif b == M : h = (r - g) / C + 4
	return [round(h % 6 * 60), s , l]

def hsl2rgb(h, s, l):
	C = (255 - abs(2*l - 255)) * s
	X = C * (1 - abs((h / 60) % 2 - 1))
	if   h < 60  : r, g, b = C, X, 0
	elif h < 120 : r, g, b = X, C, 0
	elif h < 180 : r, g, b = 0, C, X
	elif h < 240 : r, g, b = 0, X, C
	elif h < 300 : r, g, b = X, 0, C
	elif h < 360 : r, g, b = C, 0, X

	m = l - C/2
	return [r + m, g + m, b + m]

def hsla2rgba(h, s, l, a):
	return hsl2rgb(h, s, l) + [a]

def rgba2hsla(r, g, b, a):
	return rgb2hsl(r, g, b) + [a]




import os, lib.file, subprocess

__JSEXT__ = '%s.js'
__INTRO__ = __JSEXT__ % 'intro'
__OUTRO__ = __JSEXT__ % 'outro'
__INDEX__ = __JSEXT__ % 'index'

def build(root, ugly = '-uno', src = 'src', out = 'min'):

	result = os.path.join(root, out);

	if not os.path.exists(result) : os.makedirs(result)

	data = os.path.join(root, src)
	for l in os.listdir(data):

		curr = os.path.join(data, l)
		if not os.path.isdir(curr) : continue

		path = os.path.join(result, __JSEXT__ % l)

		with open(path, 'w') as f:

			def fhandle(path):
				with open(path, 'r') as g:
					f.write(os.linesep)
					f.write('/* ' + g.name + ' */')
					f.write(os.linesep)
					f.write(os.linesep)
					lib.file.read(g, f.write)
					f.write(os.linesep)

			def dhandle(d):

				el = set(os.listdir(d))

				intro = __INTRO__ in el
				outro = __OUTRO__ in el
				index = __INDEX__ in el

				if intro : el.remove(__INTRO__)
				if outro : el.remove(__OUTRO__)
				if index : el.remove(__INDEX__)

				if intro : fhandle(__INTRO__)

				for e in sorted(el):
					path = os.path.join(d, e)

					if   os.path.isdir(path)        : dhandle(path)
					elif f and os.path.isfile(path) : fhandle(path)

				if outro : fhandle(__OUTRO__)


			dhandle(curr)

		if ugly != '-uno' : uglify(path)


def uglify(path,  dest = None):
	if dest is None : dest = path
	subprocess.call(['uglifyjs', path, '-o', dest, '-m', '-c'])



import lib.bower

def register(bowername, gitendpoint, force = False):
	lib.bower.register(bowername, gitendpoint, force = force)

def unregister(bowername, force = False):
	lib.bower.unregister(bowername, force = force)




import os, lib.sys

IGNORE = '.hgignore'

def do(action, *args, **kwargs):
	lib.sys.call(['hg', action] + list(args), stddefault = None, **kwargs)

def clone(*args, **kwargs):
	do('clone', *args, **kwargs)

def pull(*args, **kwargs):
	do('pull', *args, **kwargs)

def push(*args, **kwargs):
	do('push', *args, **kwargs)

def commit(*args, **kwargs):
	do('commit', *args, **kwargs)

def add(*args, **kwargs):
	do('add', *args, **kwargs)

def status(*args, **kwargs):
	do('status', *args, **kwargs)

def diff(*args, **kwargs):
	do('diff', *args, **kwargs)

def tag(*args, **kwargs):
	do('tag', *args, **kwargs)

def parents(*args, **kwargs):
	do('parents', *args, **kwargs)




def one ( iterator ) :

	try :
		iterator = iter( iterator )
		_ , first = next( iterator )
		return all( first == rest for _ , rest in iterator )

	except StopIteration :
		return True


def select ( source , keys ) :

	return { key : source[key] for key in keys if key in source }




import os, functools

def walk(s, d = None, f = print):
	"""
		Recursively walks a directory and applies different
		callback functions depending if the item is a file
		or a directory.
	"""
	if d is None : d = functools.partial(walk, d = d, f = f)
	for e in sorted(os.listdir(s)):
		path = os.path.join(s, e)

		if   d and os.path.isdir(path)  : d(path)
		elif f and os.path.isfile(path) : f(path)


def makedirs(*dirs):
	"""
		Creates all directories given by `dirs` as well as their parents.
	"""

	for d in dirs :
		if not os.path.exists(d) : os.makedirs(d)


def size(*dirs):
	"""
		Computes the size of a list of directories.
	"""
	# http://stackoverflow.com/a/1392549/1582182
	total_size = 0

	for d in dirs :
		for dirpath, dirnames, filenames in os.walk(d):
			for f in filenames:
				fp = os.path.join(dirpath, f)
				total_size += os.path.getsize(fp)

	return total_size


class cd (object) :

	"""
		http://stackoverflow.com/a/13197763/1582182
		Context manager for changing the current working directory
	"""

	def __init__(self, cwd):
		self.cwd = cwd

	def __enter__(self):
		self.old = os.getcwd()
		os.chdir(self.cwd)

	def __exit__(self, etype, value, traceback):
		os.chdir(self.old)



import sak.navigator

PROVIDER = "www.capitainetrain.com"

def search(fr, to, go, ret = None):
	query = "https://%s/search/%s/%s/%s" % (PROVIDER, fr, to, go)
	if ret is not None : query += "/%s" % ret
	sak.navigator.open(query)




def rgba2hex(r, g, b, a = None):
	if a is None : return rgb2hex(r, g, b)
	return '#%02x%02x%02x%02x' % tuple(map(round, [r, g, b, a]))

def rgb2hex(r, g, b):
	return '#%02x%02x%02x' % tuple(map(round, [r, g, b]))

def hex2rgb(h):
	if len(h) < 9 : h += 'ff'
	return [int(h[i:i+2], 16) for i in range(1, 9, 2)]



def str2rgb(rgb):
	if rgb[3] == 'a':
		x = rgb[5:-1].split(',')
		r, g, b = map(int, x[:-1])
		a = int(255 * float(x[-1]))
		return [r, g, b, a]

	else:
		r, g, b = map(int, rgb[4:-1].split(','))
		return [r, g, b]


def rgb2str(r, g, b):
	return 'rgb(%d, %d, %d)' % tuple(map(round, [r, g, b]))

def rgba2str(r, g, b, a):
	return 'rgba(%d, %d, %d, %.2f)' % tuple(map(round, [r, g, b, a / 255]))

def str2hsl(hsl):
	if hsl[3] == 'a':
		x = hsl[5:-1].split(',')
		h = int(x[0])
		s = int(x[1][:-1]) / 100
		l = int(x[2][:-1]) / 100 * 255
		a = int(255 * float(x[-1]))
		return [h, s, l, a]
	else:
		x = hsl[4:-1].split(',')
		h = int(x[0])
		s = int(x[1][:-1]) / 100
		l = int(x[2][:-1]) / 100 * 255
		return [h, s, l]


def hsl2str(h, s, l):
	return 'hsl(%d, %d%%, %d%%)' % (h, round(s * 100), round(l / 255 * 100))

def hsla2str(h, s, l, a):
	return 'hsla(%d, %d%%, %d%%, %.2f)' % (h, round(s * 100), round(l / 255 * 100), a / 255)




import lib.shieldsio, lib.markdown

iwl = lib.markdown.imagewithlink
sio = lib.shieldsio


DEFAULT = sio.DEFAULT
FLAT = sio.FLAT
FLATSQUARED = sio.FLATSQUARED

JPG = sio.JPG
PNG = sio.PNG
SVG = sio.SVG


def npmlicense ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "NPM license"
	img = sio.npmlicense( packagename, style = style, fmt = fmt )
	href = "https://raw.githubusercontent.com/%s/%s/master/LICENSE" % ( username, repo )

	return iwl( title, img, href )


def npmversion ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "NPM version"
	img = sio.npmversion( packagename, style = style, fmt = fmt )
	href = "https://www.npmjs.org/package/%s" % ( packagename )

	return iwl( title, img, href )


def npmdownloads ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "NPM downloads per month"
	img = sio.npmdownloadsmonth( packagename, style = style, fmt = fmt )
	href = "https://www.npmjs.org/package/%s" % ( packagename )

	return iwl( title, img, href )


def bowerversion ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "Bower version"
	img = sio.bowerversion( packagename, style = style, fmt = fmt )
	href = "http://bower.io/search/?q=%s" % ( packagename )

	return iwl( title, img, href )


def traviscibuild ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "Build Status"
	img = sio.travis( username, repo, style = style, fmt = fmt )
	href = "https://travis-ci.org/%s/%s" % ( username, repo )

	return iwl( title, img, href )


def droneiobuild ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "Build Status"
	img = "https://drone.io/github.com/%s/%s/status.png" % ( username, repo )
	href = "https://drone.io/github.com/%s/%s/latest" % ( username, repo )

	return iwl( title, img, href )


def coverallscoverage ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "Coverage Status"
	img = sio.coveralls( username, repo, style = style, fmt = fmt )
	href = "https://coveralls.io/r/%s/%s" % ( username, repo )

	return iwl( title, img, href )


def daviddependencies ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "Dependencies Status"
	img = sio.david( username, repo, style = style, fmt = fmt )
	href = "https://david-dm.org/%s/%s#info=dependencies" % ( username, repo )

	return iwl( title, img, href )


def daviddevdependencies ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "devDependencies Status"
	img = sio.daviddev( username, repo, style = style, fmt = fmt )
	href = "https://david-dm.org/%s/%s#info=devDependencies" % ( username, repo )

	return iwl( title, img, href )


def codeclimategpa ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "Code Climate"
	img = sio.codeclimategithub( username, repo, style = style, fmt = fmt )
	href = "https://codeclimate.com/github/%s/%s" % ( username, repo )

	return iwl( title, img, href )


def githubissues ( username = None, repo = None, packagename = None, style = FLAT, fmt = SVG ) :

	title = "GitHub issues"
	img = sio.githubissues( username, repo, style = style, fmt = fmt )
	href = "https://github.com/%s/%s/issues" % ( username, repo )

	return iwl( title, img, href )
